package POEDaemon::WheelRun::OWNet;


use strict;
use warnings FATAL => 'all';
no warnings 'redefine';

use POE;
use POEDaemon;


sub states
{
	return $_[0],
	[qw(
		ownet_start
		ownet_stop
		
		ownet_childexit
		
		ownet
		ownet_stdout_line
		ownet_stderr_line
	)];
}


sub ownet_start
{
	my ($kernel, $heap) = @_[KERNEL, HEAP];
	
	
	log_enabled && logline 'event fired';
	
	
	my $prog =
	[
		'util/ownet.pl',
	];
	
	
	log_enabled && logline 'wheelrun_exec';
	
	
	$kernel->yield(wheelrun_exec =>
	{
		prog         => $prog,
		name         => 'ownet',
		stdout_event => 'ownet_stdout_line',
		stderr_event => 'ownet_stderr_line',
	});
}


sub ownet_stop
{
	my $kernel = $_[KERNEL];
	
	
	$kernel->yield(wheelrun_kill => { wheel_name => 'ownet' });
}


sub ownet_childexit
{
	my ($kernel, $heap) = @_[KERNEL, HEAP];
	
	
	unless ($heap->{shutdown})
	{
		$kernel->yield(shutdown => 'ownet child abnormally terminated');
	}
}


sub ownet
{
	my ($heap, $args) = @_[HEAP, ARG0];
	
	
	my $cmd = $args->{cmd};
	
	
	unless (defined $cmd)
	{
		log_enabled && logline 'undef cmd';
		
		return;
	}
	
	
	my $wheel = $heap->{wheel_run}->{children_by_name}->{ownet}->{wheel};
	
	return unless $wheel;
	
	
	#log_enabled && logline "put '%s'", $cmd;
	
	
	$wheel->put($cmd);
}


sub ownet_stdout_line
{
	my ($kernel, $session, $heap, $line) = @_[KERNEL, SESSION, HEAP, ARG0];
	
	
	unless ($line)
	{
		log_enabled && logline "invalid line = '%s'", $line;
		
		return;
	}
	
	
	my $output_per_target;
	
	
	$output_per_target->{tcpserver} = sprintf "OWNET %s", $line;
	
	$output_per_target->{tcpclient} = sprintf "%s ownet %s",
		POECLIENT_CMD_PREFIX,
		$line;
	
	
	#foreach my $target (qw(tcpserver tcpclient))
	#{
	#	next unless exists $heap->{$target} && exists $heap->{$target}->{connections};
	#	
	#	
	#	my $tcpserver = $target eq 'tcpserver';
	#	my $tcpclient = $target eq 'tcpclient';
	#	
	#	
	#	my $output = sprintf "%s %s",
	#		($tcpclient ? sprintf "%s ownet", POECLIENT_CMD_PREFIX : 'ownet'),
	#		$line;
	#	
	#	
	#	foreach my $wheel_id (keys %{$heap->{$target}->{connections}})
	#	{
	#		next if $tcpserver && !$heap->{$target}->{connections}->{$wheel_id}->{ownet_watch};
	#		
	#		
	#		next unless $heap->{$target}->{connections}->{$wheel_id}->{wheel};
	#		
	#		
	#		$kernel->call($session => sprintf("%s_output", $target) =>
	#		{
	#			wheel_id  => $wheel_id,
	#			output    => $output,
	#			use_flush => 1,
	#		});
	#	}
	#}
	
	
	$kernel->call($session => wheelrun_send_network_reply =>
	{
		output_per_target => $output_per_target,
		use_flush         => 1,
	});
	
	
	log_enabled && logline "ownet %s", $line;
}


sub ownet_stderr_line
{
	my ($kernel, $session, $heap, $line) = @_[KERNEL, SESSION, HEAP, ARG0];
	
	
	log_enabled && logline "line = '%s'", $line;
	
	
	return unless $line =~ /^ready$/i;
	
	
	$line = lc $line;
	
	
	my $output_per_target;
	
	
	$output_per_target->{tcpserver} = sprintf "OWNET %s", $line;
	
	$output_per_target->{tcpclient} = sprintf "%s ownet %s",
		POECLIENT_CMD_PREFIX,
		$line;
	
	
	#foreach my $target (qw(tcpserver tcpclient))
	#{
	#	next unless exists $heap->{$target} && exists $heap->{$target}->{connections};
	#	
	#	
	#	my $tcpserver = $target eq 'tcpserver';
	#	my $tcpclient = $target eq 'tcpclient';
	#	
	#	
	#	my $output = sprintf "%s %s",
	#		($tcpclient ? sprintf "%s ownet", POECLIENT_CMD_PREFIX : 'ownet'),
	#		lc $line;
	#	
	#	
	#	foreach my $wheel_id (keys %{$heap->{$target}->{connections}})
	#	{
	#		next unless $heap->{$target}->{connections}->{$wheel_id};
	#		
	#		next unless $heap->{$target}->{connections}->{$wheel_id}->{wheel};
	#		
	#		
	#		next if $tcpserver && !$heap->{$target}->{connections}->{$wheel_id}->{ownet_watch};
	#		
	#		
	#		$kernel->yield(sprintf("%s_output", $target) =>
	#		{
	#			wheel_id => $wheel_id,
	#			output   => $output,
	#		});
	#	}
	#}
	
	
	$kernel->call($session => wheelrun_send_network_reply =>
	{
		output_per_target => $output_per_target,
	});
}


1;
