/*
 * Copyright (c) 2014  Luiz Otavio O Souza <loos.br@gmail.com>
 * Copyright (c) 2014  Sulev-Madis Silber <ketas@si.pri.ee>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/event.h>


int main(int argc, char **argv)
{
	char *gpioc;
	int ch, fd, kq, nev, pin;
	struct kevent chlist, evlist;
	
	/* kqueue example program for userland GPIO interrupt notification. */
	
	gpioc = "/dev/gpioc0";
	while ((ch = getopt(argc, argv, "f:")) != -1)
	{
		switch (ch)
		{
			case 'f':
				gpioc = optarg;
				break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 1)
	{
		printf("Usage:\n");
		printf("gpio-intr-kqueue [-f /dev/gpiocX] interrupt-pin-number\n");
		exit(1);
	}
	
	pin = atoi(argv[0]);
	printf("Awaiting interrupts on pin: %d\n", pin);
	
	kq = kqueue();
	if (kq < 0)
	{
		err(1, "kqueue");
	}
	
	fd = open(gpioc, O_RDONLY);
	if (fd == -1)
	{
		err(1, "open");
	}
	
	/*
	 * Initialize our changes list with the pins we want to be notified
	 * about.
	 */
	EV_SET(&chlist, fd, EVFILT_READ, EV_ADD, 0, pin, NULL);
	
	for (;;)
	{
		/* Wait for an event. */
		nev = kevent(kq, &chlist, 1, &evlist, 1, NULL);
		if (nev < 0)
		{
			err(2, "kevent()");
		}
		if (nev == 0)
		{
			continue;
		}
		
		if (evlist.flags & EV_ERROR)
		{
			printf("error: %s\n", strerror(evlist.data));
			break;
		}
		
		printf("notified of isr on pin: %d\n", (int)evlist.data);
	}
	
	close(fd);
	close(kq);
	
	exit(0);
}
